from typing import Any

from ..request import Request
from .base import BaseFunction, api_function


class Manager(BaseFunction):
    """
    Provides controlling of the gateway/manager servers.

    .. versionadded:: 18.12
    """

    @api_function
    @classmethod
    async def status(cls):
        """
        Returns the current status of the configured API server.
        """
        rqst = Request("GET", "/manager/status")
        rqst.set_json({
            "status": "running",
        })
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def freeze(cls, force_kill: bool = False):
        """
        Freezes the configured API server.
        Any API clients will no longer be able to create new compute sessions nor
        create and modify vfolders/keypairs/etc.
        This is used to enter the maintenance mode of the server for unobtrusive
        manager and/or agent upgrades.

        :param force_kill: If set ``True``, immediately shuts down all running
            compute sessions forcibly. If not set, clients who have running compute
            session are still able to interact with them though they cannot create
            new compute sessions.
        """
        rqst = Request("PUT", "/manager/status")
        rqst.set_json({
            "status": "frozen",
            "force_kill": force_kill,
        })
        async with rqst.fetch():
            pass

    @api_function
    @classmethod
    async def unfreeze(cls):
        """
        Unfreezes the configured API server so that it resumes to normal operation.
        """
        rqst = Request("PUT", "/manager/status")
        rqst.set_json({
            "status": "running",
        })
        async with rqst.fetch():
            pass

    @api_function
    @classmethod
    async def get_announcement(cls):
        """
        Get current announcement.
        """
        rqst = Request("GET", "/manager/announcement")
        async with rqst.fetch() as resp:
            return await resp.json()

    @api_function
    @classmethod
    async def update_announcement(cls, enabled: bool = True, message: str = None):
        """
        Update (create / delete) announcement.

        :param enabled: If set ``False``, delete announcement.
        :param message: Announcement message. Required if ``enabled`` is True.
        """
        rqst = Request("POST", "/manager/announcement")
        rqst.set_json({
            "enabled": enabled,
            "message": message,
        })
        async with rqst.fetch():
            pass

    @api_function
    @classmethod
    async def scheduler_op(cls, op: str, args: Any):
        """
        Perform a scheduler operation.

        :param op: The name of scheduler operation.
        :param args: Arguments specific to the given operation.
        """
        rqst = Request("POST", "/manager/scheduler/operation")
        rqst.set_json({
            "op": op,
            "args": args,
        })
        async with rqst.fetch():
            pass



from typing import Any, Callable

import click

from ai.backend.common.types import SessionTypes

START_OPTION = [
    click.option(
        "-t",
        "--name",
        "--client-token",
        metavar="NAME",
        type=str,
        default=None,
        help="Specify a human-readable session name. If not set, a random hex string is used.",
    ),
    # job scheduling options
    click.option(
        "--type",
        metavar="SESSTYPE",
        type=click.Choice([*SessionTypes], case_sensitive=False),
        default=SessionTypes.INTERACTIVE,
        help="Either batch or interactive",
    ),
    click.option(
        "--starts-at",
        metavar="STARTS_AT",
        type=str,
        default=None,
        help="Let session to be started at a specific or relative time.",
    ),
    click.option(
        "--enqueue-only",
        is_flag=True,
        help="Enqueue the session and return immediately without waiting for its startup.",
    ),
    click.option(
        "--max-wait",
        metavar="SECONDS",
        type=int,
        default=0,
        help="The maximum duration to wait until the session starts.",
    ),
    click.option(
        "--no-reuse",
        is_flag=True,
        help="Do not reuse existing sessions but return an error.",
    ),
    click.option(
        "--callback-url",
        metavar="CALLBACK_URL",
        type=str,
        default=None,
        help="Callback URL which will be called upon sesison lifecycle events.",
    ),
    # execution environment
    click.option(
        "-e",
        "--env",
        metavar="KEY=VAL",
        type=str,
        multiple=True,
        help="Environment variable (may appear multiple times)",
    ),
    # extra options
    click.option(
        "--tag",
        type=str,
        default=None,
        help="User-defined tag string to annotate sessions.",
    ),
    # resource spec
    click.option(
        "-v",
        "--volume",
        "-m",
        "--mount",
        "mount",
        metavar="NAME[=PATH] or NAME[:PATH]",
        type=str,
        multiple=True,
        help=(
            "User-owned virtual folder names to mount. "
            "If path is not provided, virtual folder will be mounted under /home/work. "
            "When the target path is relative, it is placed under /home/work "
            "with auto-created parent directories if any. "
            "Absolute paths are mounted as-is, but it is prohibited to "
            "override the predefined Linux system directories."
        ),
    ),
    click.option(
        "--scaling-group",
        "--sgroup",
        metavar="SCALING_GROUP",
        type=str,
        default=None,
        help=(
            "The scaling group to execute session. If not specified, "
            "all available scaling groups are included in the scheduling."
        ),
    ),
    click.option(
        "-r",
        "--resources",
        metavar="KEY=VAL",
        type=str,
        multiple=True,
        help=(
            "Set computation resources used by the session "
            "(e.g: -r cpu=2 -r mem=256 -r gpu=1)."
            "1 slot of cpu/gpu represents 1 core. "
            "The unit of mem(ory) is MiB."
        ),
    ),
    click.option(
        "--cluster-size",
        metavar="NUMBER",
        type=int,
        default=1,
        help="The size of cluster in number of containers.",
    ),
    click.option(
        "--resource-opts",
        metavar="KEY=VAL",
        type=str,
        multiple=True,
        help="Resource options for creating compute session (e.g: shmem=64m)",
    ),
    # resource grouping
    click.option(
        "-d",
        "--domain",
        metavar="DOMAIN_NAME",
        default=None,
        help=(
            "Domain name where the session will be spawned. "
            "If not specified, config's domain name will be used."
        ),
    ),
    click.option(
        "-g",
        "--group",
        metavar="GROUP_NAME",
        default=None,
        help=(
            "Group name where the session is spawned. "
            "User should be a member of the group to execute the code."
        ),
    ),
]


def decorator_group(*decs) -> Callable[[Any], Any]:
    def d(f):
        for decorator in decs[::-1]:
            f = decorator(f)
        return f

    return d


def click_start_option() -> Callable[[Any], Any]:
    return decorator_group(
        *START_OPTION,
    )



from __future__ import annotations

import decimal
import json
import textwrap
from collections import defaultdict
from typing import Any, Callable, Mapping, Optional

import humanize

from ai.backend.common.types import MetricValue

from .types import AbstractOutputFormatter, FieldSpec


def format_stats(raw_stats: Optional[str], indent="") -> str:
    if raw_stats is None:
        return "(unavailable)"
    stats = json.loads(raw_stats)
    text = "\n".join(f"- {k + ': ':18s}{v}" for k, v in stats.items())
    return "\n" + textwrap.indent(text, indent)


def format_multiline(value: Any, indent_length: int) -> str:
    buf = []
    for idx, line in enumerate(str(value).strip().splitlines()):
        if idx == 0:
            buf.append(line)
        else:
            buf.append((" " * indent_length) + line)
    return "\n".join(buf)


def format_nested_dicts(value: Mapping[str, Mapping[str, Any]]) -> str:
    """
    Format a mapping from string keys to sub-mappings.
    """
    rows = []
    if not value:
        rows.append("(empty)")
    else:
        for outer_key, outer_value in value.items():
            if isinstance(outer_value, dict):
                if outer_value:
                    rows.append(f"+ {outer_key}")
                    inner_rows = format_nested_dicts(outer_value)
                    rows.append(textwrap.indent(inner_rows, prefix="  "))
                else:
                    rows.append(f"+ {outer_key}: (empty)")
            else:
                if outer_value is None:
                    rows.append(f"- {outer_key}: (null)")
                else:
                    # TODO: refactor as a formatter
                    if outer_key == "shmem":
                        rows.append(
                            f"- {outer_key}:"
                            f" {humanize.naturalsize(str(outer_value), binary=True, gnu=True)}"
                        )
                    else:
                        rows.append(f"- {outer_key}: {outer_value}")
    return "\n".join(rows)


def format_value(value: Any) -> str:
    if value is None:
        return "(null)"
    if isinstance(value, (dict, list, set)) and not value:
        return "(empty)"
    return str(value)


class OutputFormatter(AbstractOutputFormatter):
    """
    The base implementation of output formats.
    """

    def format_console(self, value: Any, field: FieldSpec) -> str:
        if value is None:
            return "(null)"
        if isinstance(value, (dict, list, set)) and not value:
            return "(empty)"
        elif isinstance(value, dict):
            return (
                "{"
                + ", ".join(f"{k}: {self.format_console(v, field)}" for k, v in value.items())
                + "}"
            )
        elif isinstance(value, (list, tuple, set)):
            return "[" + ", ".join(self.format_console(v, field) for v in value) + "]"
        return str(value)

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        if value is None:
            return None
        if isinstance(value, decimal.Decimal):
            return str(value)
        elif isinstance(value, dict):
            return {k: self.format_json(v, field) for k, v in value.items()}
        elif isinstance(value, (list, tuple)):
            return [self.format_json(v, field) for v in value]
        return value


class NestedDictOutputFormatter(OutputFormatter):
    def format_console(self, value: Any, field: FieldSpec) -> str:
        if value is None:
            return "(null)"
        value = json.loads(value)
        return format_nested_dicts(value)

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        if value is None:
            return None
        return json.loads(value)


class MiBytesOutputFormatter(OutputFormatter):
    def format_console(self, value: Any, field: FieldSpec) -> str:
        if value is not None:
            value = round(value / 2**20, 1)
        return super().format_console(value, field)

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        if value is not None:
            value = round(value / 2**20, 1)
        return super().format_json(value, field)


class SizeBytesOutputFormatter(OutputFormatter):
    def format_console(self, value: Any, field: FieldSpec) -> str:
        if value is not None:
            value = humanize.naturalsize(value, binary=True, gnu=True)
        return super().format_console(value, field)

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        if value is not None:
            value = humanize.naturalsize(value, binary=True, gnu=True)
        return super().format_json(value, field)


class SubFieldOutputFormatter(OutputFormatter):
    def __init__(self, subfield_name: str) -> None:
        self._subfield_name = subfield_name

    def format_console(self, value: Any, field: FieldSpec) -> str:
        return super().format_console(value[self._subfield_name], field)

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        return super().format_json(value[self._subfield_name], field)


class ResourceSlotFormatter(OutputFormatter):
    def format_console(self, value: Any, field: FieldSpec) -> str:
        value = json.loads(value)
        if mem := value.get("mem"):
            value["mem"] = humanize.naturalsize(mem, binary=True, gnu=True)

        return ", ".join(f"{k}:{v}" for k, v in value.items())

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        return json.loads(value)


default_output_formatter = OutputFormatter()
nested_dict_formatter = NestedDictOutputFormatter()
mibytes_output_formatter = MiBytesOutputFormatter()
resource_slot_formatter = ResourceSlotFormatter()
sizebytes_output_formatter = SizeBytesOutputFormatter()


class AgentStatFormatter(OutputFormatter):
    def format_console(self, value: Any, field: FieldSpec) -> str:
        try:
            raw_stats = json.loads(value)
        except TypeError:
            return ""

        percent_formatter = lambda metric, _: "{} %".format(metric["pct"])
        value_formatters: Mapping[str, Callable[[MetricValue, bool], str]] = {
            "bytes": lambda metric, binary: "{} / {}".format(
                humanize.naturalsize(int(metric["current"]), binary=binary, gnu=binary),
                (
                    humanize.naturalsize(int(metric["capacity"]), binary=binary, gnu=binary)
                    if metric["capacity"] is not None
                    else "(unknown)"
                ),
            ),
            "Celsius": lambda metric, _: "{:,} C".format(
                float(metric["current"]),
            ),
            "bps": lambda metric, _: "{}/s".format(
                humanize.naturalsize(float(metric["current"])),
            ),
            "pct": percent_formatter,
            "percent": percent_formatter,
            "%": percent_formatter,
        }

        def format_value(metric: MetricValue, binary: bool) -> str:
            unit_hint = metric["unit_hint"]
            formatter = value_formatters.get(
                unit_hint,
                # a fallback implementation
                lambda m, _: "{} / {} {}".format(
                    m["current"],
                    "(unknown)" if m["capacity"] is None else m["capacity"],
                    "" if unit_hint == "count" else unit_hint,
                ),
            )
            return formatter(metric, binary)

        bufs = []
        node_metric_bufs = []
        for stat_key, metric in raw_stats["node"].items():
            if stat_key == "cpu_util":
                num_cores = len(raw_stats["devices"]["cpu_util"])
                if metric["pct"] is None:
                    node_metric_bufs.append(f"{stat_key}: (calculating...) % ({num_cores} cores)")
                else:
                    node_metric_bufs.append(f"{stat_key}: {metric['pct']} % ({num_cores} cores)")
            else:
                binary = stat_key == "mem"
                node_metric_bufs.append(f"{stat_key}: {format_value(metric, binary)}")

        bufs.append(", ".join(node_metric_bufs))
        dev_metric_bufs = []
        for stat_key, per_dev_metric in raw_stats["devices"].items():
            dev_metric_bufs.append(f"+ {stat_key}")
            if stat_key == "cpu_util" and len(per_dev_metric) > 8:
                dev_metric_bufs.append(
                    "  - (per-core stats hidden for large CPUs with more than 8 cores)",
                )
            else:
                binary = stat_key == "mem"
                for dev_id, metric in per_dev_metric.items():
                    dev_metric_bufs.append(
                        f"  - {dev_id}: {format_value(metric, binary)}",
                    )
        bufs.append("\n".join(dev_metric_bufs))
        return "\n".join(bufs)

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        # TODO: improve
        return self.format_console(value, field)


class GroupListFormatter(OutputFormatter):
    def format_console(self, value: Any, field: FieldSpec) -> str:
        return ", ".join(g["name"] for g in value)

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        return value


class InlineRoutingFormatter(OutputFormatter):
    def format_console(self, value: Any, field: FieldSpec) -> str:
        count_by_status: defaultdict[int, int] = defaultdict(int)
        for route in value:
            count_by_status[route["status"]] += 1
        return ", ".join([
            f"{key} {value}" for key, value in sorted(count_by_status.items(), key=lambda kv: kv[0])
        ])

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        return value


class KernelStatFormatter(OutputFormatter):
    def format_console(self, value: Any, field: FieldSpec) -> str:
        return format_stats(value)

    def format_json(self, value: Any, field: FieldSpec) -> Any:
        return value


class NestedObjectFormatter(OutputFormatter):
    def format_json(self, value: Any, field: FieldSpec) -> Any:
        assert isinstance(value, list)
        return [
            {
                f.alt_name: f.formatter.format_json(item[f.field_name], f)
                for f in field.subfields.values()
            }
            for item in value
        ]


def _fit_multiline_in_cell(text: str, indent: str) -> str:
    if "\n" in text:
        return "\n" + textwrap.indent(text, indent)
    else:
        return text


class ContainerListFormatter(NestedObjectFormatter):
    def format_console(self, value: Any, field: FieldSpec, indent="") -> str:
        assert isinstance(value, list)
        if len(value) == 0:
            text = "(no sub-containers belonging to the session)"
        else:
            text = ""
            for item in value:
                text += f"+ {item['id']}\n"
                text += "\n".join(
                    f"  - {f.humanized_name}: "
                    f"{_fit_multiline_in_cell(f.formatter.format_console(item[f.field_name], f), '    ')}"  # noqa
                    for f in field.subfields.values()
                    if f.field_name != "id"
                )
        return textwrap.indent(text, indent)


class DependencyListFormatter(NestedObjectFormatter):
    def format_console(self, value: Any, field: FieldSpec, indent="") -> str:
        assert isinstance(value, list)
        if len(value) == 0:
            text = "(no dependency tasks)"
        else:
            text = ""
            for item in value:
                text += f"+ {item['name']} ({item['id']})\n"
                text += "\n".join(
                    f"  - {f.humanized_name}: "
                    f"{_fit_multiline_in_cell(f.formatter.format_console(item[f.field_name], f), '    ')}"  # noqa
                    for f in field.subfields.values()
                    if f.field_name not in ("id", "name")
                )
        return textwrap.indent(text, indent)



import asyncio
import json
import shlex
import sys
from typing import Dict, List, MutableMapping, Optional, Sequence, Union

import aiohttp
import click

from ai.backend.cli.main import main
from ai.backend.cli.types import ExitCode

from ..compat import asyncio_run, asyncio_run_forever
from ..config import DEFAULT_CHUNK_SIZE
from ..request import Request
from ..session import AsyncSession
from ..versioning import get_naming
from .pretty import print_error, print_fail, print_info, print_warn


class WSProxy:
    __slots__ = (
        "api_session",
        "session_name",
        "app_name",
        "args",
        "envs",
        "reader",
        "writer",
    )

    def __init__(
        self,
        api_session: AsyncSession,
        session_name: str,
        app_name: str,
        args: MutableMapping[str, Union[None, str, List[str]]],
        envs: MutableMapping[str, str],
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> None:
        self.api_session = api_session
        self.session_name = session_name
        self.app_name = app_name
        self.args = args
        self.envs = envs
        self.reader = reader
        self.writer = writer

    async def run(self) -> None:
        prefix = get_naming(self.api_session.api_version, "path")
        path = f"/stream/{prefix}/{self.session_name}/tcpproxy"
        params = {"app": self.app_name}

        if len(self.args.keys()) > 0:
            params["arguments"] = json.dumps(self.args)
        if len(self.envs.keys()) > 0:
            params["envs"] = json.dumps(self.envs)

        api_rqst = Request("GET", path, b"", params=params, content_type="application/json")
        async with api_rqst.connect_websocket() as ws:

            async def downstream() -> None:
                try:
                    async for msg in ws:
                        if msg.type == aiohttp.WSMsgType.ERROR:
                            await self.write_error(msg)
                            break
                        elif msg.type == aiohttp.WSMsgType.CLOSE:
                            if msg.data != aiohttp.WSCloseCode.OK:
                                await self.write_error(msg)
                            break
                        elif msg.type == aiohttp.WSMsgType.BINARY:
                            self.writer.write(msg.data)
                            await self.writer.drain()
                except ConnectionResetError:
                    pass  # shutting down
                except asyncio.CancelledError:
                    pass
                finally:
                    self.writer.close()
                    try:
                        await self.writer.wait_closed()
                    except (BrokenPipeError, IOError):
                        # closed
                        pass

            down_task = asyncio.create_task(downstream())
            try:
                while True:
                    chunk = await self.reader.read(DEFAULT_CHUNK_SIZE)
                    if not chunk:
                        break
                    await ws.send_bytes(chunk)
            except ConnectionResetError:
                pass  # shutting down
            except asyncio.CancelledError:
                raise
            finally:
                if not down_task.done():
                    down_task.cancel()
                    await down_task

    async def write_error(self, msg: aiohttp.WSMessage) -> None:
        if isinstance(msg.data, bytes):
            error_msg = msg.data.decode("utf8")
        else:
            error_msg = str(msg.data)
        rsp = (
            "HTTP/1.1 503 Service Unavailable\r\n"
            "Connection: Closed\r\n\r\n"
            "WebSocket reply: {}".format(error_msg)
        )
        self.writer.write(rsp.encode())
        await self.writer.drain()


class ProxyRunnerContext:
    __slots__ = (
        "session_name",
        "app_name",
        "protocol",
        "host",
        "port",
        "args",
        "envs",
        "api_session",
        "local_server",
        "exit_code",
    )

    session_name: str
    app_name: str
    protocol: str
    host: str
    port: int
    args: Dict[str, Union[None, str, List[str]]]
    envs: Dict[str, str]
    api_session: Optional[AsyncSession]
    local_server: Optional[asyncio.AbstractServer]
    exit_code: int

    def __init__(
        self,
        host: str,
        port: int,
        session_name: str,
        app_name: str,
        *,
        protocol: str = "tcp",
        args: Sequence[str] = None,
        envs: Sequence[str] = None,
    ) -> None:
        self.host = host
        self.port = port
        self.session_name = session_name
        self.app_name = app_name
        self.protocol = protocol

        self.api_session = None
        self.local_server = None
        self.exit_code = 0

        self.args, self.envs = {}, {}
        if args is not None and len(args) > 0:
            for argline in args:
                tokens = []
                for token in shlex.shlex(argline, punctuation_chars=True):
                    kv = token.split("=", maxsplit=1)
                    if len(kv) == 1:
                        tokens.append(shlex.split(token)[0])
                    else:
                        tokens.append(kv[0])
                        tokens.append(shlex.split(kv[1])[0])

                if len(tokens) == 1:
                    self.args[tokens[0]] = None
                elif len(tokens) == 2:
                    self.args[tokens[0]] = tokens[1]
                else:
                    self.args[tokens[0]] = tokens[1:]
        if envs is not None and len(envs) > 0:
            for envline in envs:
                split = envline.strip().split("=", maxsplit=2)
                if len(split) == 2:
                    self.envs[split[0]] = split[1]
                else:
                    self.envs[split[0]] = ""

    async def handle_connection(
        self,
        reader: asyncio.StreamReader,
        writer: asyncio.StreamWriter,
    ) -> None:
        assert self.api_session is not None
        p = WSProxy(
            self.api_session,
            self.session_name,
            self.app_name,
            self.args,
            self.envs,
            reader,
            writer,
        )
        try:
            await p.run()
        except asyncio.CancelledError:
            pass
        except Exception as e:
            print_error(e)

    async def __aenter__(self) -> None:
        self.exit_code = 0
        self.api_session = AsyncSession()
        await self.api_session.__aenter__()

        user_url_template = "{protocol}://{host}:{port}"
        try:
            compute_session = self.api_session.ComputeSession(self.session_name)
            all_apps = await compute_session.stream_app_info()
            for app_info in all_apps:
                if app_info["name"] == self.app_name:
                    if "url_template" in app_info.keys():
                        user_url_template = app_info["url_template"]
                    break
            else:
                print_fail(f'The app "{self.app_name}" is not supported by the session.')
                self.exit_code = 1
                return

            self.local_server = await asyncio.start_server(
                self.handle_connection, self.host, self.port
            )
            user_url = user_url_template.format(
                protocol=self.protocol,
                host=self.host,
                port=self.port,
            )
            print_info(
                'A local proxy to the application "{0}" '.format(self.app_name)
                + 'provided by the session "{0}" '.format(self.session_name)
                + "is available at:\n{0}".format(user_url),
            )
            if self.host == "0.0.0.0":
                print_warn(
                    'NOTE: Replace "0.0.0.0" with the actual hostname you use '
                    "to connect with the CLI app proxy."
                )
        except Exception:
            await self.api_session.__aexit__(*sys.exc_info())
            raise

    async def __aexit__(self, *exc_info) -> None:
        if self.local_server is not None:
            print_info("Shutting down....")
            self.local_server.close()
            await self.local_server.wait_closed()
        assert self.api_session is not None
        await self.api_session.__aexit__(*exc_info)
        assert self.api_session.closed
        if self.local_server is not None:
            print_info('The local proxy to "{}" has terminated.'.format(self.app_name))
        self.local_server = None


@main.command()
@click.argument("session_name", type=str, metavar="NAME")
@click.argument("app", type=str)
@click.option(
    "-b",
    "--bind",
    type=str,
    default="127.0.0.1:8080",
    metavar="[HOST:]PORT",
    help="The IP/host address and the port number to bind this proxy.",
)
@click.option(
    "--arg",
    type=str,
    multiple=True,
    metavar='"--option <value>"',
    help="Add additional argument when starting service.",
)
@click.option(
    "-e",
    "--env",
    type=str,
    multiple=True,
    metavar='"ENVNAME=envvalue"',
    help="Add additional environment variable when starting service.",
)
def app(session_name, app, bind, arg, env):
    """
    Run a local proxy to a service provided by Backend.AI compute sessions.

    The type of proxy depends on the app definition: plain TCP or HTTP.

    \b
    SESSID: The compute session ID.
    APP: The name of service provided by the given session.
    """
    bind_parts = bind.rsplit(":", maxsplit=1)
    if len(bind_parts) == 1:
        host = "127.0.0.1"
        port = int(bind_parts[0])
    elif len(bind_parts) == 2:
        host = bind_parts[0]
        port = int(bind_parts[1])
    try:
        proxy_ctx = ProxyRunnerContext(
            host,
            port,
            session_name,
            app,
            protocol="tcp",
            args=arg,
            envs=env,
        )
        asyncio_run_forever(proxy_ctx)
        sys.exit(proxy_ctx.exit_code)
    except Exception as e:
        print_error(e)
        sys.exit(ExitCode.FAILURE)


@main.command()
@click.argument("session_name", type=str, metavar="SESSION_ID", nargs=1)
@click.argument("app_name", type=str, metavar="APP", nargs=-1)
@click.option("-l", "--list-names", is_flag=True, help="Just print all available services.")
def apps(session_name, app_name, list_names):
    """
    List available additional arguments and environment variables when starting service.

    \b
    SESSID: The compute session ID.
    APP: The name of service provided by the given session. Repeatable.
         If none provided, this will print all available services.
    """

    async def print_arguments():
        apps = []
        async with AsyncSession() as api_session:
            compute_session = api_session.ComputeSession(session_name)
            apps = await compute_session.stream_app_info()
            if len(app_name) > 0:
                apps = list(filter(lambda x: x["name"] in app_name))
        if list_names:
            print_info(
                "This session provides the following app services: {0}".format(
                    ", ".join(list(map(lambda x: x["name"], apps)))
                )
            )
            return
        for service in apps:
            has_arguments = "allowed_arguments" in service.keys()
            has_envs = "allowed_envs" in service.keys()

            if has_arguments or has_envs:
                print_info("Information for service {0}:".format(service["name"]))
                if has_arguments:
                    print("\tAvailable arguments: {0}".format(service["allowed_arguments"]))
                if has_envs:
                    print("\tAvailable environment variables: {0}".format(service["allowed_envs"]))
            else:
                print_info(
                    "Service {0} does not have customizable arguments.".format(service["name"])
                )

    try:
        asyncio_run(print_arguments())
    except Exception as e:
        print_error(e)



from typing import Any, Dict, Sequence

from ...cli.types import Undefined, undefined
from ..output.fields import keypair_fields
from ..output.types import FieldSpec, PaginatedResult
from ..pagination import fetch_paginated_result
from ..session import api_session
from ..types import set_if_set
from .base import BaseFunction, api_function

__all__ = ("KeyPair",)

_default_list_fields = (
    keypair_fields["user_id"],
    keypair_fields["access_key"],
    keypair_fields["secret_key"],
    keypair_fields["is_active"],
    keypair_fields["is_admin"],
    keypair_fields["created_at"],
)

_default_detail_fields = (
    keypair_fields["user_id"],
    keypair_fields["access_key"],
    keypair_fields["secret_key"],
    keypair_fields["is_active"],
    keypair_fields["is_admin"],
)

_default_result_fields = (
    keypair_fields["access_key"],
    keypair_fields["secret_key"],
)


class KeyPair(BaseFunction):
    """
    Provides interactions with keypairs.
    """

    def __init__(self, access_key: str):
        self.access_key = access_key

    @api_function
    @classmethod
    async def create(
        cls,
        user_id: int | str,
        is_active: bool = True,
        is_admin: bool = False,
        resource_policy: str | Undefined = undefined,
        rate_limit: int | Undefined = undefined,
        fields: Sequence[FieldSpec] = _default_result_fields,
    ) -> dict:
        """
        Creates a new keypair with the given options.
        You need an admin privilege for this operation.
        """
        uid_type = "Int!" if isinstance(user_id, int) else "String!"
        q = (
            "mutation($user_id: {0}, $input: KeyPairInput!) {{".format(uid_type)
            + "  create_keypair(user_id: $user_id, props: $input) {"
            "    ok msg keypair { $fields }"
            "  }"
            "}"
        )
        q = q.replace("$fields", " ".join(f.field_ref for f in fields))
        inputs = {
            "is_active": is_active,
            "is_admin": is_admin,
        }
        set_if_set(inputs, "resource_policy", resource_policy)
        set_if_set(inputs, "rate_limit", rate_limit)
        variables = {
            "user_id": user_id,
            "input": inputs,
        }
        data = await api_session.get().Admin._query(q, variables)
        return data["create_keypair"]

    @api_function
    @classmethod
    async def update(
        cls,
        access_key: str,
        is_active: bool | Undefined = undefined,
        is_admin: bool | Undefined = undefined,
        resource_policy: str | Undefined = undefined,
        rate_limit: int | Undefined = undefined,
    ) -> dict:
        """
        Creates a new keypair with the given options.
        You need an admin privilege for this operation.
        """
        q = (
            "mutation($access_key: String!, $input: ModifyKeyPairInput!) {"
            + "  modify_keypair(access_key: $access_key, props: $input) {    ok msg  }}"
        )
        inputs: dict[str, Any] = {}
        set_if_set(inputs, "is_active", is_active)
        set_if_set(inputs, "is_admin", is_admin)
        set_if_set(inputs, "resource_policy", resource_policy)
        set_if_set(inputs, "rate_limit", rate_limit)
        variables = {
            "access_key": access_key,
            "input": inputs,
        }
        data = await api_session.get().Admin._query(q, variables)
        return data["modify_keypair"]

    @api_function
    @classmethod
    async def delete(cls, access_key: str):
        """
        Deletes an existing keypair with given ACCESSKEY.
        """
        q = (
            "mutation($access_key: String!) {"
            "  delete_keypair(access_key: $access_key) {"
            "    ok msg"
            "  }"
            "}"
        )
        variables = {
            "access_key": access_key,
        }
        data = await api_session.get().Admin._query(q, variables)
        return data["delete_keypair"]

    @api_function
    @classmethod
    async def list(
        cls,
        user_id: int | str | None = None,
        is_active: bool | None = None,
        fields: Sequence[FieldSpec] = _default_list_fields,
    ) -> Sequence[dict]:
        """
        Lists the keypairs.
        You need an admin privilege for this operation.
        """
        if user_id is None:
            q = "query($is_active: Boolean) {  keypairs(is_active: $is_active) {    $fields  }}"
        else:
            uid_type = "Int!" if isinstance(user_id, int) else "String!"
            q = (
                "query($email: {0}, $is_active: Boolean) {{".format(uid_type)
                + "  keypairs(email: $email, is_active: $is_active) {    $fields  }}"
            )
        q = q.replace("$fields", " ".join(f.field_ref for f in fields))
        variables: Dict[str, Any] = {
            "is_active": is_active,
        }
        if user_id is not None:
            variables["email"] = user_id
        data = await api_session.get().Admin._query(q, variables)
        return data["keypairs"]

    @api_function
    @classmethod
    async def paginated_list(
        cls,
        is_active: bool | None = None,
        domain_name: str | None = None,
        *,
        user_id: str | None = None,
        fields: Sequence[FieldSpec] = _default_list_fields,
        page_offset: int = 0,
        page_size: int = 20,
        filter: str | None = None,
        order: str | None = None,
    ) -> PaginatedResult[dict]:
        """
        Lists the keypairs.
        You need an admin privilege for this operation.
        """
        variables = {
            "is_active": (is_active, "Boolean"),
            "domain_name": (domain_name, "String"),
            "filter": (filter, "String"),
            "order": (order, "String"),
        }
        if user_id is not None:
            variables["email"] = (user_id, "String")
        return await fetch_paginated_result(
            "keypair_list",
            variables,
            fields,
            page_offset=page_offset,
            page_size=page_size,
        )

    @api_function
    async def info(self, fields: Sequence[FieldSpec] = _default_detail_fields) -> dict:
        """
        Returns the keypair's information such as resource limits.

        :param fields: Additional per-agent query fields to fetch.

        .. versionadded:: 18.12
        """
        q = "query {  keypair {    $fields  }}"
        q = q.replace("$fields", " ".join(f.field_ref for f in fields))
        data = await api_session.get().Admin._query(q)
        return data["keypair"]

    @api_function
    @classmethod
    async def activate(cls, access_key: str) -> dict:
        """
        Activates this keypair.
        You need an admin privilege for this operation.
        """
        q = (
            "mutation($access_key: String!, $input: ModifyKeyPairInput!) {"
            + "  modify_keypair(access_key: $access_key, props: $input) {    ok msg  }}"
        )
        variables = {
            "access_key": access_key,
            "input": {
                "is_active": True,
                "is_admin": None,
                "resource_policy": None,
                "rate_limit": None,
            },
        }
        data = await api_session.get().Admin._query(q, variables)
        return data["modify_keypair"]

    @api_function
    @classmethod
    async def deactivate(cls, access_key: str) -> dict:
        """
        Deactivates this keypair.
        Deactivated keypairs cannot make any API requests
        unless activated again by an administrator.
        You need an admin privilege for this operation.
        """
        q = (
            "mutation($access_key: String!, $input: ModifyKeyPairInput!) {"
            + "  modify_keypair(access_key: $access_key, props: $input) {    ok msg  }}"
        )
        variables = {
            "access_key": access_key,
            "input": {
                "is_active": False,
                "is_admin": None,
                "resource_policy": None,
                "rate_limit": None,
            },
        }
        data = await api_session.get().Admin._query(q, variables)
        return data["modify_keypair"]



import json
import sys

import click

from ai.backend.cli.interaction import ask_yn
from ai.backend.cli.params import OptionalType
from ai.backend.cli.types import ExitCode, Undefined, undefined
from ai.backend.common.types import VFolderHostPermission

from ...func.keypair_resource_policy import (
    _default_detail_fields,
    _default_list_fields,
)
from ...session import Session
from ..extensions import pass_ctx_obj
from ..pretty import print_info
from ..types import CLIContext

# from ai.backend.client.output.fields import keypair_resource_policy_fields
from . import admin


@admin.group()
def keypair_resource_policy() -> None:
    """
    KeyPair resource policy administration commands.
    """


@keypair_resource_policy.command()
@pass_ctx_obj
@click.argument("name", type=str)
def info(ctx: CLIContext, name: str) -> None:
    """
    Show details about a keypair resource policy. When `name` option is omitted, the
    resource policy for the current access_key will be returned.
    """
    with Session() as session:
        try:
            rp = session.KeypairResourcePolicy(session.config.access_key)
            item = rp.info(name)
            ctx.output.print_item(item, _default_detail_fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@keypair_resource_policy.command()
@pass_ctx_obj
def list(ctx: CLIContext) -> None:
    """
    List and manage keypair resource policies.
    (admin privilege required)
    """
    with Session() as session:
        try:
            items = session.KeypairResourcePolicy.list()
            ctx.output.print_list(items, _default_list_fields)
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@keypair_resource_policy.command()
@pass_ctx_obj
@click.argument("name", type=str, default=None, metavar="NAME")
@click.option(
    "--default-for-unspecified",
    type=str,
    default="UNLIMITED",
    help="Default behavior for unspecified resources: LIMITED, UNLIMITED",
)
@click.option(
    "--total-resource-slots",
    type=str,
    default="{}",
    help="Set total resource slots.",
)
@click.option(
    "--max-session-lifetime",
    type=int,
    default=0,
    help="Maximum lifetime to keep session alive.",
)
@click.option(
    "--max-concurrent-sessions",
    type=int,
    default=30,
    help="Number of maximum concurrent sessions.",
)
@click.option(
    "--max-concurrent-sftp-sessions",
    type=int,
    default=30,
    help="Number of maximum concurrent SFTP sessions.",
)
@click.option(
    "--max-containers-per-session",
    type=int,
    default=1,
    help="Number of maximum containers per session.",
)
@click.option(
    "--idle-timeout",
    type=int,
    default=1800,
    help="The maximum period of time allowed for kernels to wait further requests.",
)
@click.option(
    "--vfolder-host-perms",
    "--vfolder-host-permissions",
    "--vfhost-perms",
    "--allowed-vfolder-hosts",  # legacy name
    type=str,
    default=json.dumps({
        "local:volume1": [perm.value for perm in VFolderHostPermission],
    }),
    help=(
        "Allowed virtual folder hosts and permissions for them. It must be JSON string (e.g:"
        ' --vfolder-host-perms=\'{"HOST_NAME": ["create-vfolder", "modify-vfolder"]}\')'
    ),
)
def add(
    ctx: CLIContext,
    name: str,
    default_for_unspecified: str,
    total_resource_slots: str,  # JSON string
    max_session_lifetime: int,
    max_concurrent_sessions: int,
    max_concurrent_sftp_sessions: int,
    max_containers_per_session: int,
    idle_timeout: int,
    vfolder_host_perms: str,  # JSON string
) -> None:
    """
    Add a new keypair resource policy.

    NAME: NAME of a new keypair resource policy.
    """
    with Session() as session:
        try:
            data = session.KeypairResourcePolicy.create(
                name,
                default_for_unspecified=default_for_unspecified,
                total_resource_slots=total_resource_slots,
                max_session_lifetime=max_session_lifetime,
                max_concurrent_sessions=max_concurrent_sessions,
                max_concurrent_sftp_sessions=max_concurrent_sftp_sessions,
                max_containers_per_session=max_containers_per_session,
                idle_timeout=idle_timeout,
                vfolder_host_perms=vfolder_host_perms,
            )
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="resource_policy",
                action_name="add",
            )
            sys.exit(ExitCode.FAILURE)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="resource_policy",
                action_name="add",
            )
            sys.exit(ExitCode.FAILURE)
        ctx.output.print_mutation_result(
            data,
            item_name="resource_policy",
        )


@keypair_resource_policy.command()
@pass_ctx_obj
@click.argument("name", type=str, default=None, metavar="NAME")
@click.option(
    "--default-for-unspecified",
    type=OptionalType(str),
    default=undefined,
    help="Default behavior for unspecified resources: LIMITED, UNLIMITED",
)
@click.option(
    "--total-resource-slots",
    type=OptionalType(str),
    default=undefined,
    help="Set total resource slots.",
)
@click.option(
    "--max-session-lifetime",
    type=OptionalType(int),
    default=undefined,
    help="Maximum lifetime to keep session alive.",
)
@click.option(
    "--max-concurrent-sessions",
    type=OptionalType(int),
    default=undefined,
    help="Number of maximum concurrent sessions.",
)
@click.option(
    "--max-concurrent-sftp-sessions",
    type=OptionalType(int),
    default=undefined,
    help="Number of maximum concurrent SFTP sessions.",
)
@click.option(
    "--max-containers-per-session",
    type=OptionalType(int),
    default=undefined,
    help="Number of maximum containers per session.",
)
@click.option(
    "--idle-timeout",
    type=OptionalType(int),
    default=undefined,
    help="The maximum period of time allowed for kernels to wait further requests.",
)
@click.option(
    "--vfolder-host-perms",
    "--vfolder-host-permissions",
    "--vfhost-perms",
    "--allowed-vfolder-hosts",  # legacy name
    type=OptionalType(str),
    default=undefined,
    help=(
        "Allowed virtual folder hosts. It must be JSON string (e.g:"
        ' --vfolder-host-perms=\'{"HOST_NAME": ["create-vfolder", "modify-vfolder"]}\')'
    ),
)
def update(
    ctx: CLIContext,
    name: str,
    default_for_unspecified: str | Undefined,
    total_resource_slots: str | Undefined,  # JSON string
    max_session_lifetime: int | Undefined,
    max_concurrent_sessions: int | Undefined,
    max_concurrent_sftp_sessions: int | Undefined,
    max_containers_per_session: int | Undefined,
    idle_timeout: int | Undefined,
    vfolder_host_perms: str | Undefined,  # JSON string
) -> None:
    """
    Update an existing keypair resource policy.

    NAME: NAME of a keypair resource policy to update.
    """
    with Session() as session:
        try:
            data = session.KeypairResourcePolicy.update(
                name,
                default_for_unspecified=default_for_unspecified,
                total_resource_slots=total_resource_slots,
                max_session_lifetime=max_session_lifetime,
                max_concurrent_sessions=max_concurrent_sessions,
                max_concurrent_sftp_sessions=max_concurrent_sftp_sessions,
                max_containers_per_session=max_containers_per_session,
                idle_timeout=idle_timeout,
                vfolder_host_perms=vfolder_host_perms,
            )
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="resource_policy",
                action_name="update",
            )
            sys.exit(ExitCode.FAILURE)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="resource_policy",
                action_name="update",
            )
            sys.exit(ExitCode.FAILURE)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                "name": name,
            },
        )


@keypair_resource_policy.command()
@pass_ctx_obj
@click.argument("name", type=str, metavar="NAME")
def delete(ctx: CLIContext, name: str) -> None:
    """
    Delete a keypair resource policy.

    NAME: NAME of a keypair resource policy to delete.
    """
    with Session() as session:
        if not ask_yn():
            print_info("Cancelled.")
            sys.exit(ExitCode.FAILURE)
        try:
            data = session.KeypairResourcePolicy.delete(name)
        except Exception as e:
            ctx.output.print_mutation_error(
                e,
                item_name="resource_policy",
                action_name="deletion",
            )
            sys.exit(ExitCode.FAILURE)
        if not data["ok"]:
            ctx.output.print_mutation_error(
                msg=data["msg"],
                item_name="resource_policy",
                action_name="deletion",
            )
            sys.exit(ExitCode.FAILURE)
        ctx.output.print_mutation_result(
            data,
            extra_info={
                "name": name,
            },
        )



import json
import sys
from typing import Literal, Optional, Sequence
from uuid import UUID

import click

from ai.backend.cli.main import main
from ai.backend.cli.types import ExitCode
from ai.backend.client.cli.session.execute import prepare_env_arg, prepare_resource_arg
from ai.backend.client.compat import asyncio_run
from ai.backend.client.session import AsyncSession, Session
from ai.backend.common.arch import DEFAULT_IMAGE_ARCH
from ai.backend.common.types import ClusterMode

from ..output.fields import routing_fields, service_fields
from ..output.types import FieldSpec
from .extensions import pass_ctx_obj
from .pretty import ProgressViewer, print_done, print_fail, print_warn
from .types import CLIContext

_default_detail_fields: Sequence[FieldSpec] = (
    service_fields["endpoint_id"],
    service_fields["name"],
    service_fields["image"],
    service_fields["desired_session_count"],
    service_fields["routings"],
    service_fields["session_owner"],
    service_fields["url"],
    service_fields["open_to_public"],
)


_default_routing_fields: Sequence[FieldSpec] = (
    routing_fields["status"],
    routing_fields["traffic_ratio"],
    routing_fields["session"],
    routing_fields["endpoint"],
)


def get_service_id(session: Session, name_or_id: str):
    try:
        session.Service(name_or_id).info()
        return name_or_id
    except Exception:
        services = session.Service.list(name=name_or_id)
        try:
            return services[0]["id"]
        except Exception as e:
            if isinstance(e, KeyError) or isinstance(e, IndexError):
                raise RuntimeError(f"Service {name_or_id} not found")
            else:
                raise e


@main.group()
def service():
    """Set of service operations"""


@service.command()
@pass_ctx_obj
@click.option("--filter", "filter_", default=None, help="Set the query filter expression.")
@click.option("--order", default=None, help="Set the query ordering expression.")
@click.option("--offset", default=0, help="The index of the current page start for pagination.")
@click.option("--limit", type=int, default=None, help="The page size for pagination.")
def list(ctx: CLIContext, filter_, order, offset, limit):
    """
    List the service endpoints.
    """

    with Session() as session:
        try:
            fetch_func = lambda pg_offset, pg_size: session.Service.paginated_list(
                page_offset=pg_offset,
                page_size=pg_size,
                filter=filter_,
                order=order,
            )
            ctx.output.print_paginated_list(
                fetch_func,
                initial_page_offset=offset,
                page_size=limit,
            )
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@service.command()
@pass_ctx_obj
@click.argument("service_name_or_id", metavar="SERVICE_NAME_OR_ID", type=str)
def info(ctx: CLIContext, service_name_or_id: str):
    """
    Display the detail of a service endpoint with its backing inference session.

    \b
    ENDPOINT: The endpoint ID
    """
    with Session() as session:
        try:
            service_id = get_service_id(session, service_name_or_id)
            result = session.Service.detail(service_id, fields=_default_detail_fields)
            routes = result["routings"]
            ctx.output.print_item(
                result,
                _default_detail_fields,
            )
            print()
            for route in routes:
                print(f"Route {route['routing_id']}: ")
                ctx.output.print_item(
                    route,
                    _default_routing_fields,
                )
                print()
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@service.command()
@pass_ctx_obj
@click.argument("image", metavar="IMAGE", type=str)
@click.argument("model_name_or_id", metavar="MODEL_NAME_OR_ID", type=str)
@click.argument("initial_session_count", metavar="COUNT", type=int)
@click.option("-t", "--name", metavar="NAME", type=str, default=None)
@click.option("--model-version", metavar="VERSION", type=str, default=None)
@click.option("--model-mount-destination", metavar="PATH", type=str, default="/models")
# execution environment
@click.option(
    "-e",
    "--env",
    metavar="KEY=VAL",
    type=str,
    multiple=True,
    help="Environment variable (may appear multiple times)",
)
# extra options
@click.option(
    "--bootstrap-script",
    metavar="PATH",
    type=click.File("r"),
    default=None,
    help="A user-defined script to execute on startup.",
)
@click.option(
    "-c",
    "--startup-command",
    metavar="COMMAND",
    default=None,
    help="Set the command to execute for batch-type sessions.",
)
@click.option(
    "-r",
    "--resources",
    metavar="KEY=VAL",
    type=str,
    multiple=True,
    help=(
        "Set computation resources used by the session "
        "(e.g: -r cpu=2 -r mem=256 -r gpu=1)."
        "1 slot of cpu/gpu represents 1 core. "
        "The unit of mem(ory) is MiB."
    ),
)
@click.option(
    "--resource-opts",
    metavar="KEY=VAL",
    type=str,
    multiple=True,
    help="Resource options for creating compute session (e.g: shmem=64m)",
)
@click.option(
    "--cluster-size",
    metavar="NUMBER",
    type=int,
    default=1,
    help="The size of cluster in number of containers.",
)
@click.option(
    "--cluster-mode",
    metavar="MODE",
    type=click.Choice([*ClusterMode], case_sensitive=False),
    default=ClusterMode.SINGLE_NODE,
    help="The mode of clustering.",
)
@click.option("-d", "--domain", type=str, default="default")
@click.option("-p", "--project", type=str, default="default")
# extra options
@click.option(
    "--bootstrap-script",
    metavar="PATH",
    type=click.File("r"),
    default=None,
    help="A user-defined script to execute on startup.",
)
# extra options
@click.option("--tag", type=str, default=None, help="User-defined tag string to annotate sessions.")
@click.option(
    "--arch",
    "--architecture",
    "architecture",
    metavar="ARCH_NAME",
    type=str,
    default=DEFAULT_IMAGE_ARCH,
    help="Architecture of the image to use.",
)
@click.option(
    "--scaling-group",
    "--sgroup",
    type=str,
    default="default",
    help=(
        "The scaling group to execute session. If not specified, "
        "all available scaling groups are included in the scheduling."
    ),
)
@click.option(
    "-o",
    "--owner",
    "--owner-access-key",
    metavar="ACCESS_KEY",
    default=None,
    help="Set the owner of the target session explicitly.",
)
@click.option(
    "--public",
    "--expose-to-public",
    is_flag=True,
    help=(
        "Visibility of API Endpoint which serves inference workload."
        "If set to true, no authentication will be required to access the endpoint."
    ),
)
def create(
    ctx: CLIContext,
    image: str,
    model_name_or_id: str,
    initial_session_count: int,
    *,
    name: Optional[str],
    model_version: Optional[str],
    model_mount_destination: Optional[str],
    env: Sequence[str],
    startup_command: Optional[str],
    resources: Sequence[str],
    resource_opts: Sequence[str],
    cluster_size: int,
    cluster_mode: Literal["single-node", "multi-node"],
    domain: Optional[str],
    project: Optional[str],
    bootstrap_script: Optional[str],
    tag: Optional[str],
    architecture: Optional[str],
    scaling_group: Optional[str],
    owner: Optional[str],
    public: bool,
):
    """
    Create a service endpoint with a backing inference session.

    \b
    MODEL_ID: The model ID

    """
    envs = prepare_env_arg(env)
    parsed_resources = prepare_resource_arg(resources)
    parsed_resource_opts = prepare_resource_arg(resource_opts)
    body = {
        "service_name": name,
        "model_version": model_version,
        "envs": envs,
        "startup_command": startup_command,
        "resources": parsed_resources,
        "resource_opts": parsed_resource_opts,
        "cluster_size": cluster_size,
        "cluster_mode": cluster_mode,
        "bootstrap_script": bootstrap_script,
        "tag": tag,
        "architecture": architecture,
        "scaling_group": scaling_group,
        "expose_to_public": public,
    }
    if model_mount_destination:
        body["model_mount_destination"] = model_mount_destination
    if domain:
        body["domain_name"] = domain
    if project:
        body["group_name"] = project
    if owner:
        body["owner_access_key"] = owner

    with Session() as session:
        try:
            result = session.Service.create(
                image,
                model_name_or_id,
                initial_session_count,
                **body,
            )
            ctx.output.print_item(
                result,
                [*service_fields.values()],
            )
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@service.command()
@pass_ctx_obj
@click.argument("image", metavar="IMAGE", type=str)
@click.argument("model_name_or_id", metavar="MODEL_NAME_OR_ID", type=str)
@click.option("-t", "--name", metavar="NAME", type=str, default=None)
@click.option("--model-version", metavar="VERSION", type=int, default=1)
@click.option("--model-mount-destination", metavar="PATH", type=str, default="/models")
# execution environment
@click.option(
    "-e",
    "--env",
    metavar="KEY=VAL",
    type=str,
    multiple=True,
    help="Environment variable (may appear multiple times)",
)
# extra options
@click.option(
    "--bootstrap-script",
    metavar="PATH",
    type=click.File("r"),
    default=None,
    help="A user-defined script to execute on startup.",
)
@click.option(
    "-c",
    "--startup-command",
    metavar="COMMAND",
    default=None,
    help="Set the command to execute for batch-type sessions.",
)
@click.option(
    "-r",
    "--resources",
    metavar="KEY=VAL",
    type=str,
    multiple=True,
    help=(
        "Set computation resources used by the session "
        "(e.g: -r cpu=2 -r mem=256 -r gpu=1)."
        "1 slot of cpu/gpu represents 1 core. "
        "The unit of mem(ory) is MiB."
    ),
)
@click.option(
    "--resource-opts",
    metavar="KEY=VAL",
    type=str,
    multiple=True,
    help="Resource options for creating compute session (e.g: shmem=64m)",
)
@click.option(
    "--cluster-size",
    metavar="NUMBER",
    type=int,
    default=1,
    help="The size of cluster in number of containers.",
)
@click.option(
    "--cluster-mode",
    metavar="MODE",
    type=click.Choice([*ClusterMode], case_sensitive=False),
    default=ClusterMode.SINGLE_NODE,
    help="The mode of clustering.",
)
@click.option("-d", "--domain", type=str, default="default")
@click.option("-p", "--project", type=str, default="default")
# extra options
@click.option(
    "--bootstrap-script",
    metavar="PATH",
    type=click.File("r"),
    default=None,
    help="A user-defined script to execute on startup.",
)
# extra options
@click.option("--tag", type=str, default=None, help="User-defined tag string to annotate sessions.")
@click.option(
    "--arch",
    "--architecture",
    "architecture",
    metavar="ARCH_NAME",
    type=str,
    default=DEFAULT_IMAGE_ARCH,
    help="Architecture of the image to use.",
)
@click.option(
    "--scaling-group",
    "--sgroup",
    type=str,
    default="default",
    help=(
        "The scaling group to execute session. If not specified, "
        "all available scaling groups are included in the scheduling."
    ),
)
@click.option(
    "-o",
    "--owner",
    "--owner-access-key",
    metavar="ACCESS_KEY",
    default=None,
    help="Set the owner of the target session explicitly.",
)
@click.option(
    "--public",
    "--expose-to-public",
    is_flag=True,
    help=(
        "Visibility of API Endpoint which serves inference workload."
        "If set to true, no authentication will be required to access the endpoint."
    ),
)
def try_start(
    ctx: CLIContext,
    image: str,
    model_name_or_id: str,
    *,
    name: Optional[str],
    model_version: int,
    model_mount_destination: Optional[str],
    env: Sequence[str],
    startup_command: Optional[str],
    resources: Sequence[str],
    resource_opts: Sequence[str],
    cluster_size: int,
    cluster_mode: Literal["single-node", "multi-node"],
    domain: Optional[str],
    project: Optional[str],
    bootstrap_script: Optional[str],
    tag: Optional[str],
    architecture: Optional[str],
    scaling_group: Optional[str],
    owner: Optional[str],
    public: bool,
):
    """
    Tries to create a model service session and return whether the server has successfully started or not.

    \b
    MODEL_ID: The model ID

    """
    envs = prepare_env_arg(env)
    parsed_resources = prepare_resource_arg(resources)
    parsed_resource_opts = prepare_resource_arg(resource_opts)
    body = {
        "service_name": name,
        "model_version": model_version,
        "envs": envs,
        "startup_command": startup_command,
        "resources": parsed_resources,
        "resource_opts": parsed_resource_opts,
        "cluster_size": cluster_size,
        "cluster_mode": cluster_mode,
        "bootstrap_script": bootstrap_script,
        "tag": tag,
        "architecture": architecture,
        "scaling_group": scaling_group,
        "expose_to_public": public,
    }
    if model_mount_destination:
        body["model_mount_destination"] = model_mount_destination
    if domain:
        body["domain_name"] = domain
    if project:
        body["group_name"] = project
    if owner:
        body["owner_access_key"] = owner

    with Session() as session:
        try:
            result = session.Service.try_start(
                image,
                model_name_or_id,
                **body,
            )
            ctx.output.print_item(
                result,
                [*service_fields.values()],
            )
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)

    async def try_start_tracker(bgtask_id):
        async with AsyncSession() as session:
            try:
                bgtask = session.BackgroundTask(bgtask_id)
                completion_msg_func = lambda: print_done("Model service validation started.")
                async with (
                    bgtask.listen_events() as response,
                    ProgressViewer("Starting the session...") as viewer,
                ):
                    async for ev in response:
                        data = json.loads(ev.data)
                        if ev.event == "bgtask_updated":
                            print(data["message"])
                            if viewer.tqdm is None:
                                pbar = await viewer.to_tqdm()
                            else:
                                pbar.total = data["total_progress"]
                                pbar.update(data["current_progress"] - pbar.n)
                        elif ev.event == "bgtask_failed":
                            error_msg = data["message"]
                            completion_msg_func = lambda: print_fail(
                                f"Error during the operation: {error_msg}",
                            )
                        elif ev.event == "bgtask_cancelled":
                            completion_msg_func = lambda: print_warn(
                                "The operation has been cancelled in the middle. "
                                "(This may be due to server shutdown.)",
                            )
            finally:
                completion_msg_func()

    asyncio_run(try_start_tracker(result["task_id"]))


@service.command()
@pass_ctx_obj
@click.argument("service_name_or_id", metavar="SERVICE_NAME_OR_ID", type=str)
def rm(ctx: CLIContext, service_name_or_id):
    """
    Remove the service endpoint.

    \b
    SERVICE_ID: The endpoint ID"""
    with Session() as session:
        try:
            service_id = get_service_id(session, service_name_or_id)
            session.Service(service_id).delete()
            print_done("Removed.")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@service.command()
@pass_ctx_obj
@click.argument("service_name_or_id", metavar="SERVICE_NAME_OR_ID", type=str)
def sync(ctx: CLIContext, service_name_or_id: str):
    """
    Sync route status with AppProxy.

    \b
    SERVICE_ID: The endpoint ID
    """
    with Session() as session:
        try:
            service_id = get_service_id(session, service_name_or_id)
            session.Service(service_id).sync()
            print_done("Done.")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@service.command()
@pass_ctx_obj
@click.argument("service_name_or_id", metavar="SERVICE_NAME_OR_ID", type=str)
@click.argument("target_count", metavar="COUNT", type=int)
def scale(ctx: CLIContext, service_name_or_id: str, target_count: int):
    """
    Start or resume the service endpoint to handle the incoming traffic.

    \b
    SERVICE_ID: The endpoint ID
    COUNT: Number of desired sessions in an endpoint
    """
    with Session() as session:
        try:
            service_id = get_service_id(session, service_name_or_id)
            session.Service(service_id).scale(target_count)
            print_done("Triggered scaling.")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@service.command()
@pass_ctx_obj
@click.argument("service_name_or_id", metavar="SERVICE_NAME_OR_ID", type=str)
@click.argument("duration", metavar="DURATION", type=str)
@click.option("-q", "--quiet", is_flag=True)
def generate_token(ctx: CLIContext, service_name_or_id: str, duration: str, quiet: bool):
    """
    Generate an API token to communicate with inference endpoint.

    \b
    SERVICE_ID: The endpoint ID
    DURATION: Total amount of time for token to be alive
              in short format (e.g. 3d, 2h, ...)
    """
    with Session() as session:
        try:
            service_id = get_service_id(session, service_name_or_id)
            resp = session.Service(service_id).generate_api_token(duration)
            if quiet:
                print(resp["token"])
            else:
                print_done(f"Generated API token {resp['token']}")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@service.command()
@pass_ctx_obj
@click.argument("service_name_or_id", metavar="SERVICE_NAME_OR_ID", type=str)
def get_endpoint(ctx: CLIContext, service_name_or_id: str):
    """
    Returns API Endpoint URL of the service.

    \b
    ENDPOINT: The endpoint ID
    """
    with Session() as session:
        try:
            service_id = get_service_id(session, service_name_or_id)
            result = session.Service.detail(service_id, fields=_default_detail_fields)
            print(result["url"])
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@service.command()
@pass_ctx_obj
@click.argument("service_name_or_id", metavar="SERVICE_NAME_OR_ID", type=str)
@click.argument("route_id", metavar="ROUTE_ID", type=str)
@click.argument("ratio", metavar="RATIO", type=float)
def update_traffic_ratio(ctx: CLIContext, service_name_or_id: str, route_id: str, ratio: float):
    """
    Update traffic ratio of single route.

    \b
    SERVICE_ID: The endpoint ID
    ROUTE_ID: The route ID
    RATIO: Target traffic ratio
    """
    with Session() as session:
        try:
            service_id = get_service_id(session, service_name_or_id)
            session.Service(service_id).update_traffic_ratio(UUID(route_id), ratio)
            print_done("Done.")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)


@service.command()
@pass_ctx_obj
@click.argument("service_name_or_id", metavar="SERVICE_NAME_OR_ID", type=str)
@click.argument("route_id", metavar="ROUTE_ID", type=str)
@click.argument("ratio", metavar="RATIO", type=float)
def downscale_route(ctx: CLIContext, service_name_or_id: str, route_id: str):
    """
    Destroy route and its associated session and
    decrement desired session count of endpoint

    \b
    SERVICE_ID: The endpoint ID
    ROUTE_ID: Route ID to destroy
    """
    with Session() as session:
        try:
            service_id = get_service_id(session, service_name_or_id)
            session.Service(service_id).downscale_single_route(UUID(route_id))
            print_done("Done.")
        except Exception as e:
            ctx.output.print_error(e)
            sys.exit(ExitCode.FAILURE)



from __future__ import annotations

import textwrap
from typing import Any, Dict, Final, Sequence, Tuple, TypeVar

from .exceptions import BackendAPIVersionError
from .output.types import FieldSpec, PaginatedResult
from .session import api_session

MAX_PAGE_SIZE: Final = 100
MIN_PAGE_SIZE: Final = 1

T = TypeVar("T")


async def execute_paginated_query(
    root_field: str,
    variables: Dict[str, Tuple[Any, str]],
    fields: Sequence[FieldSpec],
    *,
    limit: int,
    offset: int,
) -> PaginatedResult:
    if limit > MAX_PAGE_SIZE:
        raise ValueError(f"The page size cannot exceed {MAX_PAGE_SIZE}")
    if limit < MIN_PAGE_SIZE:
        raise ValueError(f"The page size cannot be less than {MIN_PAGE_SIZE}")
    query = """
    query($limit:Int!, $offset:Int!, $var_decls) {
      $root_field(
          limit:$limit, offset:$offset, $var_args) {
        items { $fields }
        total_count
      }
    }"""
    query = query.replace("$root_field", root_field)
    query = query.replace("$fields", " ".join(f.field_ref for f in fields))
    query = query.replace(
        "$var_decls",
        ", ".join(f"${key}: {value[1]}" for key, value in variables.items()),
    )
    query = query.replace(
        "$var_args",
        ", ".join(f"{key}:${key}" for key in variables.keys()),
    )
    query = textwrap.dedent(query).strip()
    var_values = {key: value[0] for key, value in variables.items()}
    var_values["limit"] = limit
    var_values["offset"] = offset
    data = await api_session.get().Admin._query(query, var_values)
    return PaginatedResult(
        total_count=data[root_field]["total_count"],
        items=data[root_field]["items"],
        fields=fields,
    )


async def fetch_paginated_result(
    root_field: str,
    variables: Dict[str, Tuple[Any, str]],
    fields: Sequence[FieldSpec],
    *,
    page_offset: int,
    page_size: int,
) -> PaginatedResult:
    if page_size > MAX_PAGE_SIZE:
        raise ValueError(f"The page size cannot exceed {MAX_PAGE_SIZE}")
    if page_size < MIN_PAGE_SIZE:
        raise ValueError(f"The page size cannot be less than {MIN_PAGE_SIZE}")
    if api_session.get().api_version < (6, "20210815"):
        if variables["filter"][0] is not None or variables["order"][0] is not None:
            raise BackendAPIVersionError(
                "filter and order arguments for paginated lists require v6.20210815 or later.",
            )
        # should remove to work with older managers
        variables.pop("filter")
        variables.pop("order")
    result = await execute_paginated_query(
        root_field,
        variables,
        fields,
        limit=page_size,
        offset=page_offset,
    )
    return result



from __future__ import annotations

from .formatters import (
    AgentStatFormatter,
    ContainerListFormatter,
    DependencyListFormatter,
    GroupListFormatter,
    InlineRoutingFormatter,
    KernelStatFormatter,
    SubFieldOutputFormatter,
    mibytes_output_formatter,
    nested_dict_formatter,
    resource_slot_formatter,
    sizebytes_output_formatter,
)
from .types import FieldSet, FieldSpec

container_fields = FieldSet([
    FieldSpec("id", "Kernel ID", alt_name="kernel_id"),
    FieldSpec("cluster_role"),
    FieldSpec("cluster_idx"),
    FieldSpec("cluster_hostname"),
    FieldSpec("session_id", "Session ID"),
    FieldSpec("image"),
    FieldSpec("registry"),
    FieldSpec("status"),
    FieldSpec("status_info"),
    FieldSpec("status_data", formatter=nested_dict_formatter),
    FieldSpec("status_changed"),
    FieldSpec("agent"),
    FieldSpec("container_id"),
    FieldSpec("resource_opts", formatter=nested_dict_formatter),
    FieldSpec("occupied_slots", formatter=resource_slot_formatter),
    FieldSpec("preopen_ports", "Preopen Ports"),
    FieldSpec("live_stat", formatter=KernelStatFormatter()),
    FieldSpec("last_stat", formatter=KernelStatFormatter()),
])


agent_fields = FieldSet([
    FieldSpec("id"),
    FieldSpec("status"),
    FieldSpec("status_changed"),
    FieldSpec("region"),
    FieldSpec("architecture"),
    FieldSpec("scaling_group"),
    FieldSpec("schedulable"),
    FieldSpec("available_slots", formatter=resource_slot_formatter),
    FieldSpec("occupied_slots", formatter=resource_slot_formatter),
    FieldSpec("addr"),
    FieldSpec("first_contact"),
    FieldSpec("lost_at"),
    FieldSpec("live_stat", formatter=AgentStatFormatter()),
    FieldSpec("version"),
    FieldSpec("compute_plugins"),
    FieldSpec("hardware_metadata", formatter=nested_dict_formatter),
    FieldSpec("compute_containers", subfields=container_fields, formatter=ContainerListFormatter()),
    FieldSpec("local_config", formatter=nested_dict_formatter),
    # legacy fields
    FieldSpec("cpu_cur_pct", "CPU Usage (%)"),
    FieldSpec("mem_cur_bytes", "Used Memory (MiB)", formatter=mibytes_output_formatter),
])

domain_fields = FieldSet([
    FieldSpec("name"),
    FieldSpec("description"),
    FieldSpec("is_active"),
    FieldSpec("created_at"),
    FieldSpec("total_resource_slots", formatter=resource_slot_formatter),
    FieldSpec("allowed_vfolder_hosts"),
    FieldSpec("allowed_docker_registries"),
    FieldSpec("integration_id"),
])


group_fields = FieldSet([
    FieldSpec("id"),
    FieldSpec("name"),
    FieldSpec("description"),
    FieldSpec("is_active"),
    FieldSpec("created_at"),
    FieldSpec("domain_name"),
    FieldSpec("total_resource_slots", formatter=resource_slot_formatter),
    FieldSpec("allowed_vfolder_hosts"),
    FieldSpec("integration_id"),
])


image_fields = FieldSet([
    FieldSpec("name"),
    FieldSpec("registry"),
    FieldSpec("architecture"),
    FieldSpec("tag"),
    FieldSpec("digest"),
    FieldSpec("size_bytes", formatter=sizebytes_output_formatter),
    FieldSpec("aliases"),
])


keypair_fields = FieldSet([
    FieldSpec("user_id", "Email"),
    FieldSpec(
        "user_info { full_name }",
        "Full Name",
        alt_name="full_name",
        formatter=SubFieldOutputFormatter("full_name"),
    ),
    FieldSpec("projects"),
    FieldSpec("access_key"),
    FieldSpec("secret_key"),
    FieldSpec("is_active"),
    FieldSpec("is_admin"),
    FieldSpec("created_at"),
    FieldSpec("modified_at"),
    FieldSpec("last_used"),
    FieldSpec("resource_policy"),
    FieldSpec("rate_limit"),
    FieldSpec("concurrency_used"),
    FieldSpec("ssh_public_key"),
    FieldSpec("ssh_private_key"),
    FieldSpec("dotfiles"),
    FieldSpec("bootstrap_script"),
])


keypair_resource_policy_fields = FieldSet([
    FieldSpec("name"),
    FieldSpec("created_at"),
    FieldSpec("total_resource_slots"),
    FieldSpec("max_concurrent_sessions"),  # formerly concurrency_limit
    FieldSpec("idle_timeout"),
    FieldSpec("max_containers_per_session"),
    FieldSpec("allowed_vfolder_hosts"),
])


scaling_group_fields = FieldSet([
    FieldSpec("name"),
    FieldSpec("description"),
    FieldSpec("is_active"),
    FieldSpec("is_public"),
    FieldSpec("created_at"),
    FieldSpec("driver"),
    FieldSpec("driver_opts", formatter=nested_dict_formatter),
    FieldSpec("scheduler"),
    FieldSpec("scheduler_opts", formatter=nested_dict_formatter),
    FieldSpec("use_host_network"),
    FieldSpec("wsproxy_addr"),
    FieldSpec("wsproxy_api_token"),
])


session_fields = FieldSet([
    FieldSpec("id", "Session ID", alt_name="session_id"),
    FieldSpec("main_kernel_id", "Main Kernel ID"),
    FieldSpec("tag"),
    FieldSpec("name"),
    FieldSpec("type"),
    FieldSpec("image"),
    FieldSpec("registry"),
    FieldSpec("cluster_template"),
    FieldSpec("cluster_mode"),
    FieldSpec("cluster_size"),
    FieldSpec("domain_name"),
    FieldSpec("group_name", "Project/Group"),
    FieldSpec("group_id"),
    FieldSpec("agent_ids"),
    FieldSpec("user_email"),
    FieldSpec("user_id"),
    FieldSpec("access_key", "Owner Access Key"),
    FieldSpec("created_user_email"),
    FieldSpec("created_user_id"),
    FieldSpec("status"),
    FieldSpec("status_info"),
    FieldSpec("status_data", formatter=nested_dict_formatter),
    FieldSpec("status_changed", "Last Updated"),
    FieldSpec("created_at"),
    FieldSpec("terminated_at"),
    FieldSpec("starts_at"),
    FieldSpec("scheduled_at"),
    FieldSpec("startup_command"),
    FieldSpec("result"),
    FieldSpec("resource_opts", formatter=nested_dict_formatter),
    FieldSpec("scaling_group"),
    FieldSpec("service_ports", formatter=nested_dict_formatter),
    FieldSpec("mounts"),
    FieldSpec("vfolder_mounts"),
    FieldSpec("occupying_slots", formatter=resource_slot_formatter),
    FieldSpec(
        "containers",
        subfields=container_fields,
        formatter=ContainerListFormatter(),
    ),
    FieldSpec(
        "dependencies { name id }",
        formatter=DependencyListFormatter(),
    ),
    FieldSpec("abusing_reports"),
    FieldSpec("idle_checks"),
])

session_fields_v5 = FieldSet([
    FieldSpec(
        "containers",
        subfields=FieldSet([
            FieldSpec("id", "Kernel ID", alt_name="kernel_id"),
            FieldSpec("session_id", "Session ID"),
            FieldSpec("role"),
            FieldSpec("agent"),
            FieldSpec("image"),
            FieldSpec("status"),
            FieldSpec("status_info"),
            FieldSpec("status_data", formatter=nested_dict_formatter),
            FieldSpec("status_changed"),
            FieldSpec("occupied_slots", formatter=resource_slot_formatter),
            FieldSpec("live_stat", formatter=KernelStatFormatter()),
            FieldSpec("last_stat", formatter=KernelStatFormatter()),
        ]),
        formatter=ContainerListFormatter(),
    ),
])


storage_fields = FieldSet([
    FieldSpec("id"),
    FieldSpec("backend"),
    FieldSpec("fsprefix"),
    FieldSpec("path"),
    FieldSpec("capabilities"),
    FieldSpec("hardware_metadata", formatter=nested_dict_formatter),
    FieldSpec("performance_metric", formatter=nested_dict_formatter),
    FieldSpec("usage", formatter=nested_dict_formatter),
])


user_fields = FieldSet([
    FieldSpec("uuid"),
    FieldSpec("username"),
    FieldSpec("email"),
    # password is not queriable!
    FieldSpec("need_password_change"),
    FieldSpec("full_name"),
    FieldSpec("description"),
    FieldSpec("is_active"),
    FieldSpec("status"),
    FieldSpec("status_info"),
    FieldSpec("created_at"),
    FieldSpec("modified_at"),
    FieldSpec("domain_name"),
    FieldSpec("role"),
    FieldSpec("groups { id name }", formatter=GroupListFormatter()),
    FieldSpec("allowed_client_ip"),
    FieldSpec("totp_activated"),
    FieldSpec("sudo_session_enabled"),
    FieldSpec("main_access_key"),
])


vfolder_fields = FieldSet([
    FieldSpec("id"),
    FieldSpec("host"),
    FieldSpec("name"),
    FieldSpec("user", alt_name="user_id"),
    FieldSpec("group", alt_name="group_id"),
    FieldSpec("creator"),
    FieldSpec("status"),
    FieldSpec("unmanaged_path"),
    FieldSpec("usage_mode"),
    FieldSpec("permission"),
    FieldSpec("ownership_type"),
    FieldSpec("max_files"),
    FieldSpec("max_size"),
    FieldSpec("created_at"),
    FieldSpec("last_used"),
    FieldSpec("num_files"),
    FieldSpec("cur_size"),
    FieldSpec("cloneable"),
])


permission_fields = FieldSet([
    FieldSpec("vfolder_host_permission_list"),
])


service_fields = FieldSet([
    FieldSpec("endpoint_id"),
    FieldSpec("image"),
    FieldSpec("domain"),
    FieldSpec("project"),
    FieldSpec("resource_group"),
    FieldSpec("resource_slots", formatter=nested_dict_formatter),
    FieldSpec("url"),
    FieldSpec("model"),
    FieldSpec("model_mount_destiation"),
    FieldSpec("created_user"),
    FieldSpec("session_owner"),
    FieldSpec("tag"),
    FieldSpec("startup_command"),
    FieldSpec("bootstrap_script"),
    FieldSpec("callback_url"),
    FieldSpec("environ", formatter=nested_dict_formatter),
    FieldSpec("name"),
    FieldSpec("resource_opts", formatter=nested_dict_formatter),
    FieldSpec("desired_session_count"),
    FieldSpec("cluster_mode"),
    FieldSpec("cluster_size"),
    FieldSpec("open_to_public"),
    FieldSpec(
        "routings { routing_id session status traffic_ratio }",
        formatter=InlineRoutingFormatter(),
    ),
])


routing_fields = FieldSet([
    FieldSpec("routing_id"),
    FieldSpec("status"),
    FieldSpec("endpoint"),
    FieldSpec("session"),
    FieldSpec("traffic_ratio"),
])


quota_scope_fields = FieldSet([
    FieldSpec("usage_bytes", formatter=sizebytes_output_formatter),
    FieldSpec("usage_count"),
    FieldSpec("hard_limit_bytes", formatter=sizebytes_output_formatter),
    FieldSpec("quota_scope_id"),
    FieldSpec("storage_host_name"),
])
